@@ -23,11 +23,9 @@
 #include <cstring>
 #include <iostream>
 #include <map>
-#include <set>
 
 #include <Poco/Path.h>
 #include <Poco/Process.h>
-#include <Poco/StringTokenizer.h>
 #include <Poco/Thread.h>
 #include <Poco/Util/Application.h>
 
@@ -43,19 +41,24 @@
 #include "security.h"
 
 using Poco::Process;
-using Poco::StringTokenizer;
 using Poco::Thread;
+#ifndef KIT_IN_PROCESS
 using Poco::Util::Application;
+#endif
 
+#ifndef KIT_IN_PROCESS
 static bool NoCapsForKit = false;
+#endif
 static bool DisplayVersion = false;
 static std::string UnitTestLibrary;
-static std::atomic<unsigned> ForkCounter( 0 );
+static std::atomic<unsigned> ForkCounter(0);
 
 static std::map<Process::PID, std::string> childJails;
 
+#ifndef KIT_IN_PROCESS
 int ClientPortNumber = DEFAULT_CLIENT_PORT_NUMBER;
 int MasterPortNumber = DEFAULT_MASTER_PORT_NUMBER;
+#endif
 
 /// Dispatcher class to demultiplex requests from
 /// WSD and handles them.
@@ -91,26 +94,33 @@ public:
         }
 
         LOG_INF("ForKit command: [" << message << "].");
-        StringTokenizer tokens(message, " ", StringTokenizer::TOK_IGNORE_EMPTY | StringTokenizer::TOK_TRIM);
-
-        if (tokens[0] == "spawn" && tokens.count() == 2)
+        try
         {
-            const auto count = std::stoi(tokens[1]);
-            if (count > 0)
+            std::vector<std::string> tokens = LOOLProtocol::tokenize(message);
+            if (tokens.size() == 2 && tokens[0] == "spawn")
             {
-                LOG_INF("Spawning " << tokens[1] << ' ' << (count == 1 ? "child" : "children") << " per request.");
-                ForkCounter = count;
-            }
-            else
-            {
-                LOG_WRN("Cannot spawn " << tokens[1] << " children as requested.");
+                const auto count = std::stoi(tokens[1]);
+                if (count > 0)
+                {
+                    LOG_INF("Setting to spawn " << tokens[1] << " child" << (count == 1 ? "" : "ren") << " per request.");
+                    ForkCounter = count;
+                }
+                else
+                {
+                    LOG_WRN("Cannot spawn " << tokens[1] << " children as requested.");
+                }
             }
         }
+        catch (const std::exception& exc)
+        {
+            LOG_ERR("Error while processing forkit request [" << message << "]: " << exc.what());
+        }
 
         return true;
     }
 };
 
+#ifndef KIT_IN_PROCESS
 static bool haveCapability(cap_value_t capability)
 {
     cap_t caps = cap_get_proc();
@@ -179,25 +189,35 @@ static bool haveCorrectCapabilities()
 
     return result;
 }
+#endif
 
 /// Check if some previously forked kids have died.
 static void cleanupChildren()
 {
+    std::vector<std::string> jails;
     Process::PID exitedChildPid;
     int status;
+    // Reap quickly without doing slow cleanup so WSD can spawn more rapidly.
     while ((exitedChildPid = waitpid(-1, &status, WUNTRACED | WNOHANG)) > 0)
     {
-        if (childJails.find(exitedChildPid) != childJails.end())
+        const auto it = childJails.find(exitedChildPid);
+        if (it != childJails.end())
         {
-            LOG_INF("Child " << exitedChildPid << " has exited, removing its jail '" << childJails[exitedChildPid] << "'");
-            FileUtil::removeFile(childJails[exitedChildPid], true);
-            childJails.erase(exitedChildPid);
+            LOG_INF("Child " << exitedChildPid << " has exited, removing its jail '" << it->second << "'.");
+            jails.emplace_back(it->second);
+            childJails.erase(it);
         }
         else
         {
             LOG_ERR("Unknown child " << exitedChildPid << " has exited");
         }
     }
+
+    // Now delete the jails.
+    for (const auto& path : jails)
+    {
+        FileUtil::removeFile(path, true);
+    }
 }
 
 static int createLibreOfficeKit(const std::string& childRoot,
@@ -206,17 +226,22 @@ static int createLibreOfficeKit(const std::string& childRoot,
                                 const std::string& loSubPath,
                                 bool queryVersion = false)
 {
-    LOG_DBG("Forking a loolkit process.");
+    // Generate a jail ID to be used for in the jail path.
+    const std::string jailId = Util::rng::getFilename(16);
 
-    Process::PID pid;
-    if (!(pid = fork()))
+    LOG_DBG("Forking a loolkit process with jailId: " << jailId << ".");
+
+    const Process::PID pid = fork();
+    if (!pid)
     {
         // Child
 
         // Close the pipe from loolwsd
         close(0);
 
+#ifndef KIT_IN_PROCESS
         UnitKit::get().postFork();
+#endif
 
         if (std::getenv("SLEEPKITFORDEBUGGER"))
         {
@@ -230,7 +255,11 @@ static int createLibreOfficeKit(const std::string& childRoot,
             }
         }
 
-        lokit_main(childRoot, sysTemplate, loTemplate, loSubPath, NoCapsForKit, queryVersion, DisplayVersion);
+#ifndef KIT_IN_PROCESS
+        lokit_main(childRoot, jailId, sysTemplate, loTemplate, loSubPath, NoCapsForKit, queryVersion, DisplayVersion);
+#else
+        lokit_main(childRoot, jailId, sysTemplate, loTemplate, loSubPath, true, queryVersion, DisplayVersion);
+#endif
     }
     else
     {
@@ -242,15 +271,51 @@ static int createLibreOfficeKit(const std::string& childRoot,
         else
         {
             LOG_INF("Forked kit [" << pid << "].");
-            childJails[pid] = childRoot + std::to_string(pid);
+            childJails[pid] = childRoot + jailId;
         }
 
+#ifndef KIT_IN_PROCESS
         UnitKit::get().launchedKit(pid);
+#endif
     }
 
     return pid;
 }
 
+void forkLibreOfficeKit(const std::string& childRoot,
+                        const std::string& sysTemplate,
+                        const std::string& loTemplate,
+                        const std::string& loSubPath,
+                        int limit)
+{
+    // Cleanup first, to reduce disk load.
+    cleanupChildren();
+
+#ifndef KIT_IN_PROCESS
+    (void) limit;
+#else
+    if (limit > 0)
+        ForkCounter = limit;
+#endif
+
+    if (ForkCounter > 0)
+    {
+        // Create as many as requested.
+        const size_t count = ForkCounter;
+        LOG_INF("Spawning " << count << " new child" << (count == 1 ? "." : "ren."));
+        const size_t retry = count * 2;
+        for (size_t i = 0; ForkCounter > 0 && i < retry; ++i)
+        {
+            if (ForkCounter-- <= 0 || createLibreOfficeKit(childRoot, sysTemplate, loTemplate, loSubPath) < 0)
+            {
+                LOG_ERR("Failed to create a kit process.");
+                ++ForkCounter;
+            }
+        }
+    }
+}
+
+#ifndef KIT_IN_PROCESS
 static void printArgumentHelp()
 {
     std::cout << "Usage: loolforkit [OPTION]..." << std::endl;
@@ -279,8 +344,12 @@ int main(int argc, char** argv)
         }
     }
 
+#ifndef FUZZER
     SigUtil::setFatalSignals();
     SigUtil::setTerminationSignals();
+#endif
+
+    Util::setThreadName("forkit");
 
     // Initialization
     const bool logToFile = std::getenv("LOOL_LOGFILE");
@@ -372,18 +441,43 @@ int main(int argc, char** argv)
         return Application::EXIT_USAGE;
     }
 
-    if (!UnitBase::init(UnitBase::UnitType::TYPE_KIT,
+    if (!UnitBase::init(UnitBase::UnitType::Kit,
                         UnitTestLibrary))
     {
         LOG_ERR("Failed to load kit unit test library");
         return Application::EXIT_USAGE;
     }
 
-    if (!std::getenv("LD_BIND_NOW"))
+    // Setup & check environment
+    std::string layers(
+        "xcsxcu:${BRAND_BASE_DIR}/share/registry "
+        "res:${BRAND_BASE_DIR}/share/registry "
+        "bundledext:${${BRAND_BASE_DIR}/program/lounorc:BUNDLED_EXTENSIONS_USER}/registry/com.sun.star.comp.deployment.configuration.PackageRegistryBackend/configmgr.ini "
+        "sharedext:${${BRAND_BASE_DIR}/program/lounorc:SHARED_EXTENSIONS_USER}/registry/com.sun.star.comp.deployment.configuration.PackageRegistryBackend/configmgr.ini "
+        "userext:${${BRAND_BASE_DIR}/program/lounorc:UNO_USER_PACKAGES_CACHE}/registry/com.sun.star.comp.deployment.configuration.PackageRegistryBackend/configmgr.ini "
+#if ENABLE_DEBUG // '*' denotes non-writable.
+        "user:*file://" DEBUG_ABSSRCDIR "/loolkitconfig.xcu "
+#else
+        "user:*file://" LOOLWSD_CONFIGDIR "/loolkitconfig.xcu "
+#endif
+        );
+
+    // No-caps tracing can spawn eg. glxinfo & other oddness.
+    unsetenv("DISPLAY");
+
+    ::setenv("CONFIGURATION_LAYERS", layers.c_str(),
+             1 /* override */);
+
+    if (!std::getenv("LD_BIND_NOW")) // must be set by parent.
         LOG_INF("Note: LD_BIND_NOW is not set.");
 
-    if (!haveCorrectCapabilities())
+    if (!NoCapsForKit && !haveCorrectCapabilities())
+    {
+        std::cerr << "FATAL: Capabilities are not set for the loolforkit program." << std::endl;
+        std::cerr << "Please make sure that the current partition was *not* mounted with the 'nosuid' option." << std::endl;
+        std::cerr << "If you are on SLES11, please set 'file_caps=1' as kernel boot option." << std::endl << std::endl;
         return Application::EXIT_SOFTWARE;
+    }
 
     // Initialize LoKit
     if (!globalPreinit(loTemplate))
@@ -393,9 +487,10 @@ int main(int argc, char** argv)
 
     // We must have at least one child, more are created dynamically.
     // Ask this first child to send version information to master process
-    if (createLibreOfficeKit(childRoot, sysTemplate, loTemplate, loSubPath, true) < 0)
+    Process::PID forKitPid = createLibreOfficeKit(childRoot, sysTemplate, loTemplate, loSubPath, true);
+    if (forKitPid < 0)
     {
-        Log::fatal("Failed to create a kit process.");
+        LOG_FTL("Failed to create a kit process.");
         std::_Exit(Application::EXIT_SOFTWARE);
     }
 
@@ -412,38 +507,20 @@ int main(int argc, char** argv)
             break;
         }
 
-        if (ForkCounter > 0)
-        {
-            // Create as many as requested.
-            int spawn = ForkCounter;
-            LOG_INF("Creating " << spawn << " new child.");
-            size_t newInstances = 0;
-            do
-            {
-                if (createLibreOfficeKit(childRoot, sysTemplate, loTemplate, loSubPath) < 0)
-                {
-                    LOG_ERR("Failed to create a kit process.");
-                }
-                else
-                {
-                    ++newInstances;
-                }
-            }
-            while (--spawn > 0);
-
-            // If we need to spawn more, retry later.
-            ForkCounter = (newInstances >= ForkCounter ? 0 : ForkCounter - newInstances);
-        }
-
-        // We virtually always fork when a child exits.
-        cleanupChildren();
+        forkLibreOfficeKit(childRoot, sysTemplate, loTemplate, loSubPath);
     }
 
     int returnValue = Application::EXIT_OK;
     UnitKit::get().returnValue(returnValue);
 
+#if 0
+    int status = 0;
+    waitpid(forKitPid, &status, WUNTRACED);
+#endif
+
     LOG_INF("ForKit process finished.");
     std::_Exit(returnValue);
 }
+#endif
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
