@@ -35,8 +35,7 @@ L.Control.ContextMenu = L.Control.extend({
 				   'UpdateCurIndex','RemoveTableOf',
 				   'ReplyComment', 'DeleteComment', 'DeleteAuthor', 'DeleteAllNotes'],
 
-			spreadsheet: ['MergeCells', 'SplitCells',
-						  'InsertAnnotation', 'EditAnnotation', 'DeleteNote', 'ShowNote', 'HideNote'],
+			spreadsheet: ['MergeCells', 'SplitCells', 'RecalcPivotTable'],
 
 			presentation: [],
 			drawing: []
@@ -128,23 +127,39 @@ L.Control.ContextMenu = L.Control.extend({
 					// Lets use the later for simplicity and to leverage the core translations in online
 					itemName = itemName.replace(itemName.substring('Delete All Comments by '.length), 'This Author');
 				}
+
+				switch (commandName) {
+				case 'Cut':
+					itemName = _('Internal Cut');
+					break;
+				case 'Copy':
+					itemName = _('Internal Copy');
+					break;
+				case 'Paste':
+					itemName = _('Internal Paste');
+					break;
+				}
+
 				contextMenu[item.command] = {
 					name: _(itemName)
 				};
 
 				if (item.checktype === 'checkmark') {
 					if (item.checked === 'true') {
-						contextMenu[item.command.icon] = 'checkmark';
+						contextMenu[item.command]['icon'] = 'lo-checkmark';
 					}
 				} else if (item.checktype === 'radio') {
 					if (item.checked === 'true') {
-						contextMenu[item.command.icon] = 'radio';
+						contextMenu[item.command]['icon'] = 'radio';
 					}
 				}
 
 				isLastItemText = true;
 			} else if (item.type === 'menu') {
 				itemName = item.text.replace('~', '');
+				if (itemName === 'Paste Special') {
+					itemName = _('Internal Paste Special');
+				}
 				var submenu = this._createContextMenuStructure(item);
 				// ignore submenus with all items disabled
 				if (Object.keys(submenu).length === 0) {
