@@ -44,7 +44,11 @@ L.Map.include({
 	},
 
 	getToolbarCommandValues: function (command) {
-		return this._docLayer._toolbarCommandValues[command];
+		if (this._docLayer) {
+			return this._docLayer._toolbarCommandValues[command];
+		}
+
+		return undefined;
 	},
 
 	downloadAs: function (name, format, options, id) {
@@ -52,13 +56,20 @@ L.Map.include({
 			return;
 		}
 
+		id = id || 'export'; // not any special download, simple export
+
+		if ((id === 'print' && this['wopi'].DisablePrint) ||
+		    (id === 'export' && this['wopi'].DisableExport)) {
+			this.hideBusy();
+			return;
+		}
+
 		if (format === undefined || format === null) {
 			format = '';
 		}
 		if (options === undefined || options === null) {
 			options = '';
 		}
-		id = id || -1; // not a special download
 
 		this.showBusy(_('Downloading...'), false);
 		this._socket.sendMessage('downloadas ' +
@@ -120,17 +131,7 @@ L.Map.include({
 	},
 
 	save: function(dontTerminateEdit, dontSaveIfUnmodified) {
-		var args = {
-			DontTerminateEdit: {
-				type: 'boolean',
-				value: !!dontTerminateEdit
-			},
-			DontSaveIfUnmodified: {
-				type: 'boolean',
-				value: !!dontSaveIfUnmodified
-			}
-		};
-		this.sendUnoCommand('.uno:Save', args);
+		this._socket.sendMessage('save dontTerminateEdit=' + (dontTerminateEdit ? 1 : 0) + ' dontSaveIfUnmodified=' + (dontSaveIfUnmodified ? 1 : 0));
 	},
 
 	sendUnoCommand: function (command, json) {
@@ -208,6 +209,8 @@ L.Map.include({
 						translatableContent[i].firstChild.nodeValue = translatableContent[i].firstChild.nodeValue.toLocaleString();
 					}
 
+					$('.vex-content').attr('tabindex', -1);
+					$('.vex-content').focus();
 					// workaround for https://github.com/HubSpot/vex/issues/43
 					$('.vex-overlay').css({ 'pointer-events': 'none'});
 					$('.vex').click(function() {
@@ -216,6 +219,9 @@ L.Map.include({
 					$('.vex-content').click(function(e) {
 						e.stopPropagation();
 					});
+				},
+				beforeClose: function () {
+					map.focus();
 				}
 			});
 		});
@@ -238,6 +244,7 @@ L.Map.include({
 			contentCSS: { width: w + 'px'},
 			buttons: {},
 			afterOpen: function($vexContent) {
+				map.enable(false);
 				// workaround for https://github.com/HubSpot/vex/issues/43
 				$('.vex-overlay').css({ 'pointer-events': 'none'});
 				$('.vex').click(function() {
@@ -246,6 +253,9 @@ L.Map.include({
 				$('.vex-content').click(function(e) {
 					e.stopPropagation();
 				});
+			},
+			beforeClose: function () {
+				map.enable(true);
 			}
 		});
 	}
