@@ -1,3 +1,4 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.Socket contains methods for the communication with the server
  */
@@ -5,9 +6,23 @@
 /* global _ vex $ errorMessages */
 L.Socket = L.Class.extend({
 	ProtocolVersionNumber: '0.1',
+	ReconnectCount: 0,
+
+	getParameterValue: function (s) {
+		var i = s.indexOf('=');
+		if (i === -1)
+			return undefined;
+		return s.substring(i+1);
+	},
 
 	initialize: function (map) {
+		console.debug('socket.initialize:');
 		this._map = map;
+		this._msgQueue = [];
+	},
+
+	connect: function() {
+		var map = this._map;
 		try {
 			if (map.options.permission) {
 				map.options.docParams['permission'] = map.options.permission;
@@ -22,7 +37,27 @@ L.Socket = L.Class.extend({
 			this._map.fire('error', {msg: _('Oops, there is a problem connecting to LibreOffice Online : ' + e), cmd: 'socket', kind: 'failed', id: 3});
 			return null;
 		}
-		this._msgQueue = [];
+
+		if (map.options.docParams.access_token && parseInt(map.options.docParams.access_token_ttl)) {
+			var tokenExpiryWarning = 900 * 1000; // Warn when 15 minutes remain
+			clearTimeout(this._accessTokenExpireTimeout);
+			this._accessTokenExpireTimeout = setTimeout(L.bind(this._sessionExpiredWarning, this),
+			                                            parseInt(map.options.docParams.access_token_ttl) - Date.now() - tokenExpiryWarning);
+		}
+	},
+
+	_sessionExpiredWarning: function() {
+		clearTimeout(this._accessTokenExpireTimeout);
+		var expirymsg = errorMessages.sessionexpiry;
+		if (parseInt(this._map.options.docParams.access_token_ttl) - Date.now() <= 0) {
+			expirymsg = errorMessages.sessionexpired;
+		}
+		var timerepr = $.timeago(parseInt(this._map.options.docParams.access_token_ttl)).replace(' ago', '');
+		this._map.fire('warn', {msg: expirymsg.replace('%time', timerepr)});
+
+		// If user still doesn't refresh the session, warn again periodically
+		this._accessTokenExpireTimeout = setTimeout(L.bind(this._sessionExpiredWarning, this),
+		                                            120 * 1000);
 	},
 
 	close: function () {
@@ -30,6 +65,8 @@ L.Socket = L.Class.extend({
 		this.socket.onclose = function () {};
 		this.socket.onmessage = function () {};
 		this.socket.close();
+
+		clearTimeout(this._accessTokenExpireTimeout);
 	},
 
 	connected: function() {
@@ -54,7 +91,7 @@ L.Socket = L.Class.extend({
 			if (typeof msg === 'string') {
 				L.Log.log(msg, L.OUTGOING, coords);
 				if (this._map._docLayer && this._map._docLayer._debug) {
-					console.log(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c ', 1), 'background:#fbb;color:black', 'color:red', 'color:black');
+					console.log2(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c '), 'background:#fbb;color:black', 'color:red', 'color:black');
 				}
 			}
 		}
@@ -69,7 +106,7 @@ L.Socket = L.Class.extend({
 		if (typeof msg === 'string') {
 			L.Log.log(msg, L.OUTGOING, coords);
 			if (this._map._docLayer && this._map._docLayer._debug) {
-				console.log(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c ', 1), 'background:#fbb;color:black', 'color:red', 'color:black');
+				console.log2(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c '), 'background:#fbb;color:black', 'color:red', 'color:black');
 			}
 		}
 
@@ -77,6 +114,8 @@ L.Socket = L.Class.extend({
 	},
 
 	_onSocketOpen: function () {
+		console.debug('_onSocketOpen:');
+		this._map._serverRecycling = false;
 		// Always send the protocol version number.
 		// TODO: Move the version number somewhere sensible.
 		this._doSend('loolclient ' + this.ProtocolVersionNumber);
@@ -93,6 +132,9 @@ L.Socket = L.Class.extend({
 		if (this._map._docPassword) {
 			msg += ' password=' + this._map._docPassword;
 		}
+		if (String.locale) {
+			msg += ' lang=' + String.locale;
+		}
 		if (this._map.options.renderingOptions) {
 			var options = {
 				'rendering': this._map.options.renderingOptions
@@ -100,7 +142,6 @@ L.Socket = L.Class.extend({
 			msg += ' options=' + JSON.stringify(options);
 		}
 		this._doSend(msg);
-		this._doSend('partpagerectangles');
 		for (var i = 0; i < this._msgQueue.length; i++) {
 			this._doSend(this._msgQueue[i].msg, this._msgQueue[i].coords);
 		}
@@ -134,7 +175,7 @@ L.Socket = L.Class.extend({
 		}
 
 		if (this._map._docLayer && this._map._docLayer._debug) {
-			console.log(+new Date() + ' %cINCOMING%c: ' + textMsg.concat(' ').replace(' ', '%c ', 1), 'background:#ddf;color:black', 'color:blue', 'color:black');
+			console.log2(+new Date() + ' %cINCOMING%c: ' + textMsg.concat(' ').replace(' ', '%c '), 'background:#ddf;color:black', 'color:blue', 'color:black');
 		}
 
 		var command = this.parseServerCmd(textMsg);
@@ -142,8 +183,8 @@ L.Socket = L.Class.extend({
 			// This must be the first message, unless we reconnect.
 			var loolwsdVersionObj = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
 			var h = loolwsdVersionObj.Hash;
-			if (parseInt(h,16).toString(16) === h.toLowerCase()) {
-				h = '<a target="_blank" href="https://gerrit.libreoffice.org/gitweb?p=online.git;a=log;h=' + h + '">' + h + '</a>';
+			if (parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
+				h = '<a target="_blank" href="https://hub.libreoffice.org/git-online/' + h + '">' + h + '</a>';
 				$('#loolwsd-version').html(loolwsdVersionObj.Version + ' (git hash: ' + h + ')');
 			}
 			else {
@@ -158,8 +199,8 @@ L.Socket = L.Class.extend({
 		else if (textMsg.startsWith('lokitversion ')) {
 			var lokitVersionObj = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
 			var h = lokitVersionObj.BuildId.substring(0, 7);
-			if (parseInt(h,16).toString(16) === h.toLowerCase()) {
-				h = '<a target="_blank" href="https://gerrit.libreoffice.org/gitweb?p=core.git;a=log;h=' + h + '">' + h + '</a>';
+			if (parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
+				h = '<a target="_blank" href="https://hub.libreoffice.org/git-core/' + h + '">' + h + '</a>';
 			}
 			$('#lokit-version').html(lokitVersionObj.ProductName + ' ' +
 			                         lokitVersionObj.ProductVersion + lokitVersionObj.ProductExtension.replace('.10.','-') +
@@ -192,8 +233,42 @@ L.Socket = L.Class.extend({
 			if (textMsg === 'ownertermination') {
 				msg = _('Session terminated by document owner');
 			}
-			else if (textMsg === 'shutdown') {
-				msg = _('Server shutdown for maintenance');
+			else if (textMsg === 'idle') {
+				msg = _('Session terminated due to idleness');
+			}
+			else if (textMsg === 'shuttingdown') {
+				msg = _('Server is shutting down for maintenance (auto-saving)');
+			}
+			else if (textMsg === 'recycling') {
+				msg = _('Server is recycling and will be available shortly');
+
+				this._map._active = false;
+				this._map._serverRecycling = true;
+
+				// Prevent reconnecting the world at the same time.
+				var min = 5000;
+				var max = 10000;
+				var timeoutMs = Math.floor(Math.random() * (max - min) + min);
+
+				socket = this;
+				map = this._map;
+				vex.timer = setInterval(function() {
+					if (socket.connected()) {
+						// We're connected: cancel timer and dialog.
+						clearTimeout(vex.timer);
+						if (vex.dialogID > 0) {
+							var id = vex.dialogID;
+							vex.dialogID = -1;
+							vex.close(id);
+						}
+						return;
+					}
+
+					try {
+						socket.initialize(map);
+					} catch (error) {
+					}
+				}, timeoutMs);
 			}
 
 			// Close any open dialogs first.
@@ -230,23 +305,43 @@ L.Socket = L.Class.extend({
 			$(options.appendLocation).append(options.$vex);
 			vex.setupBodyClassName(options.$vex);
 
-			// Disconnect the websocket manually
-			this.close();
+			if (textMsg !== 'shuttingdown') {
+				// Tell WOPI host about it which should handle this situation
+				this._map.fire('postMessage', {msgId: 'Session_Closed'});
+			}
 
-			// Tell WOPI host about it which should handle this situation
-			this._map.fire('postMessage', {msgId: 'Session_Closed'});
-			this._map.remove();
+			if (textMsg === 'idle' || textMsg === 'ownertermination') {
+				this._map.remove();
+			}
 
 			return;
 		}
 		else if (textMsg.startsWith('error:') && command.errorCmd === 'storage') {
+			var storageError;
 			if (command.errorKind === 'savediskfull') {
-				this._map.fire('error', {msg: errorMessages.storage.savediskfull});
+				storageError = errorMessages.storage.savediskfull;
 			}
 			else if (command.errorKind === 'savefailed') {
-				// Just warn the user
-				this._map.fire('warn', {msg: errorMessages.storage.savefailed});
+				storageError = errorMessages.storage.savefailed;
 			}
+			else if (command.errorKind === 'saveunauthorized') {
+				storageError = errorMessages.storage.saveunauthorized;
+			}
+			else if (command.errorKind === 'loadfailed') {
+				storageError = errorMessages.storage.loadfailed;
+				// Since this is a document load failure, wsd will disconnect the socket anyway,
+				// better we do it first so that another error message doesn't override this one
+				// upon socket close.
+				this._map.hideBusy();
+				this.close();
+			}
+
+			// Parse the storage url as link
+			var tmpLink = document.createElement('a');
+			tmpLink.href = this._map.options.doc;
+			// Insert the storage server address to be more friendly
+			storageError = storageError.replace('%storageserver', tmpLink.host);
+			this._map.fire('warn', {msg: storageError});
 
 			return;
 		}
@@ -256,6 +351,7 @@ L.Socket = L.Class.extend({
 				this._map.fire('error', {msg: errorMessages.diskfull});
 			}
 			else if (command.errorKind === 'unauthorized') {
+				this._map.hideBusy();
 				this._map.fire('error', {msg: errorMessages.unauthorized});
 			}
 
@@ -286,6 +382,26 @@ L.Socket = L.Class.extend({
 			} else if (errorKind.startsWith('wrongpassword')) {
 				passwordNeeded = true;
 				msg = _('Wrong password provided. Please try again.');
+			} else if (errorKind.startsWith('faileddocloading')) {
+				this._map._fatal = true;
+				this._map.fire('error', {msg: errorMessages.faileddocloading});
+			} else if (errorKind.startsWith('docunloading')) {
+				// The document is unloading. Have to wait a bit.
+				this._map._active = false;
+
+				if (this.ReconnectCount++ >= 10) {
+					clearTimeout(vex.timer);
+					return; // Give up.
+				}
+
+				map = this._map;
+				vex.timer = setInterval(function() {
+					try {
+						// Activate and cancel timer and dialogs.
+						map._activate();
+					} catch (error) {
+					}
+				}, 1000);
 			}
 
 			if (passwordNeeded) {
@@ -312,14 +428,25 @@ L.Socket = L.Class.extend({
 			textMsg = textMsg.substring(6);
 			if (command.errorKind === 'limitreached') {
 				this._map._fatal = true;
-				textMsg = errorMessages.limitreached;
-				textMsg = textMsg.replace(/%0/g, command.params[0]);
-				textMsg = textMsg.replace(/%1/g, command.params[1]);
-				textMsg = textMsg.replace(/%2/g, (typeof brandProductName !== 'undefined' ? brandProductName : 'LibreOffice Online'));
-				textMsg = textMsg.replace(/%3/g, (typeof brandProductFAQURL !== 'undefined' ? brandProductFAQURL : 'https://wiki.documentfoundation.org/Development/LibreOffice_Online'));
+				this._map._active = false; // Practically disconnected.
+
+				// Servers configured for 50 documents are not demo/development.
+				if (parseInt(command.params[0]) >= 50) {
+					textMsg = errorMessages.limitreachedprod;
+					textMsg = textMsg.replace(/%0/g, command.params[0]);
+					textMsg = textMsg.replace(/%1/g, command.params[1]);
+				}
+				else {
+					textMsg = errorMessages.limitreached;
+					textMsg = textMsg.replace(/%0/g, command.params[0]);
+					textMsg = textMsg.replace(/%1/g, command.params[1]);
+					textMsg = textMsg.replace(/%2/g, (typeof brandProductName !== 'undefined' ? brandProductName : 'LibreOffice Online'));
+					textMsg = textMsg.replace(/%3/g, (typeof brandProductFAQURL !== 'undefined' ? brandProductFAQURL : 'https://wiki.documentfoundation.org/Development/LibreOffice_Online'));
+				}
 			}
 			else if (command.errorKind === 'serviceunavailable') {
 				this._map._fatal = true;
+				this._map._active = false; // Practically disconnected.
 				textMsg = errorMessages.serviceunavailable;
 			}
 			this._map.fire('error', {msg: textMsg});
@@ -335,6 +462,16 @@ L.Socket = L.Class.extend({
 		else if (textMsg.startsWith('statusindicator:')) {
 			//FIXME: We should get statusindicator when saving too, no?
 			this._map.showBusy(_('Connecting...'), false);
+			if (textMsg.startsWith('statusindicator: ready')) {
+				// We're connected: cancel timer and dialog.
+				this.ReconnectCount = 0;
+				clearTimeout(vex.timer);
+				if (vex.dialogID > 0) {
+					var id = vex.dialogID;
+					vex.dialogID = -1;
+					vex.close(id);
+				}
+			}
 		}
 		else if (!textMsg.startsWith('tile:') && !textMsg.startsWith('renderfont:')) {
 			// log the tile msg separately as we need the tile coordinates
@@ -413,7 +550,7 @@ L.Socket = L.Class.extend({
 			this._map.fire('doclayerinit');
 		} else if (textMsg.startsWith('status:') && this._reconnecting) {
 			// we are reconnecting ...
-			this._reconecting = false;
+			this._reconnecting = false;
 			this._map._docLayer._onMessage('invalidatetiles: EMPTY', null);
 			this._map.fire('statusindicator', {statusType: 'reconnected'});
 			this._map.setPermission(this._map.options.permission);
@@ -441,11 +578,13 @@ L.Socket = L.Class.extend({
 	},
 
 	_onSocketError: function () {
+		console.debug('_onSocketError:');
 		this._map.hideBusy();
 		// Let onclose (_onSocketClose) report errors.
 	},
 
 	_onSocketClose: function (e) {
+		console.debug('_onSocketClose:');
 		var isActive = this._map._active;
 		this._map.hideBusy();
 		this._map._active = false;
@@ -454,12 +593,18 @@ L.Socket = L.Class.extend({
 			this._map._docLayer.removeAllViews();
 		}
 
-		if (isActive) {
+		if (isActive && this._reconnecting) {
+			// Don't show this before first transparently trying to reconnect.
 			this._map.fire('error', {msg: _('Well, this is embarrassing, we cannot connect to your document. Please try again.'), cmd: 'socket', kind: 'closed', id: 4});
 		}
 
 		// Reset wopi's app loaded so that reconnecting again informs outerframe about initialization again
 		this._map['wopi'].resetAppLoaded();
+
+		if (!this._reconnecting) {
+			this._reconnecting = true;
+			this._map._activate();
+		}
 	},
 
 	parseServerCmd: function (msg) {
@@ -507,9 +652,6 @@ L.Socket = L.Class.extend({
 				// remove newline characters
 				command.type = tokens[i].substring(5).replace(/(\r\n|\n|\r)/gm, '');
 			}
-			else if (tokens[i].substring(0, 9) === 'prefetch=') {
-				command.preFetch = tokens[i].substring(9);
-			}
 			else if (tokens[i].substring(0, 4) === 'cmd=') {
 				command.errorCmd = tokens[i].substring(4);
 			}
@@ -532,7 +674,10 @@ L.Socket = L.Class.extend({
 				command.port = tokens[i].substring(5);
 			}
 			else if (tokens[i].substring(0, 5) === 'font=') {
-				command.font = window.decodeURIComponent(tokens[i].substring(5));
+				command.font = tokens[i].substring(5);
+			}
+			else if (tokens[i].substring(0, 5) === 'char=') {
+				command.char = tokens[i].substring(5);
 			}
 			else if (tokens[i].substring(0, 7) === 'viewid=') {
 				command.viewid = tokens[i].substring(7);
@@ -546,6 +691,9 @@ L.Socket = L.Class.extend({
 			else if (tokens[i].substring(0, 12) === 'rendercount=') {
 				command.rendercount = parseInt(tokens[i].substring(12));
 			}
+			else if (tokens[i].startsWith('hash=')) {
+				command.hash = this.getParameterValue(tokens[i]);
+			}
 		}
 		if (command.tileWidth && command.tileHeight && this._map._docLayer) {
 			var defaultZoom = this._map.options.zoom;
