@@ -1,3 +1,4 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.GridLayer is used as base class for grid-like layers like TileLayer.
  */
@@ -362,13 +363,14 @@ L.GridLayer = L.Layer.extend({
 		this._tileHeightTwips = Math.round(this.options.tileHeightTwips * factor);
 	},
 
-	_updateMaxBounds: function (sizeChanged) {
+	_updateMaxBounds: function (sizeChanged, extraSize, options) {
 		if (this._docWidthTwips === undefined || this._docHeightTwips === undefined) {
 			return;
 		}
 		var docPixelLimits = new L.Point(this._docWidthTwips / this.options.tileWidthTwips,
-										 this._docHeightTwips / this.options.tileHeightTwips);
-		docPixelLimits = docPixelLimits.multiplyBy(this._tileSize);
+			this._docHeightTwips / this.options.tileHeightTwips);
+		docPixelLimits = extraSize ? docPixelLimits.multiplyBy(this._tileSize).add(extraSize) :
+			docPixelLimits.multiplyBy(this._tileSize);
 
 		var scale = this._map.getZoomScale(this._map.getZoom(), 10);
 		var topLeft = new L.Point(0, 0);
@@ -378,12 +380,13 @@ L.GridLayer = L.Layer.extend({
 
 		if (this._documentInfo === '' || sizeChanged) {
 			// we just got the first status so we need to center the document
-			this._map.setMaxBounds(new L.LatLngBounds(topLeft, bottomRight));
+			this._map.setMaxBounds(new L.LatLngBounds(topLeft, bottomRight), options);
 		}
 
 		var scrollPixelLimits = new L.Point(this._docWidthTwips / this._tileWidthTwips,
-										 this._docHeightTwips / this._tileHeightTwips);
-		scrollPixelLimits = scrollPixelLimits.multiplyBy(this._tileSize);
+			this._docHeightTwips / this._tileHeightTwips);
+		scrollPixelLimits = extraSize ? scrollPixelLimits.multiplyBy(this._tileSize).add(extraSize.multiplyBy(scale)) :
+			scrollPixelLimits.multiplyBy(this._tileSize);
 		this._docPixelSize = {x: scrollPixelLimits.x, y: scrollPixelLimits.y};
 		this._map.fire('docsize', {x: scrollPixelLimits.x, y: scrollPixelLimits.y});
 	},
@@ -535,7 +538,13 @@ L.GridLayer = L.Layer.extend({
 				// so we're able to cancel the previous requests that are being processed
 				this._map._socket.sendMessage('canceltiles');
 				for (key in this._tiles) {
-					if (!this._tiles[key].loaded) {
+					// When _invalidCount > 0 the tile has been invalidated, however the new tile content
+					// has not yet been fetched and because of `canceltiles` message it will never be
+					// so we need to remove the tile, or when the tile is back inside the visible area
+					// its content would be the old invalidated one;
+					// example: a tile is invalidated but a sudden scroll to the cell cursor position causes
+					// to move the tile out of the visible area before the new content is fetched
+					if (!this._tiles[key].loaded || this._tiles[key]._invalidCount > 0) {
 						L.DomUtil.remove(this._tiles[key].el);
 						delete this._tiles[key];
 						if (this._debug) {
@@ -608,7 +617,13 @@ L.GridLayer = L.Layer.extend({
 				this._map._socket.sendMessage('canceltiles');
 				for (key in this._tiles) {
 					tile = this._tiles[key];
-					if (!tile.loaded) {
+					// When _invalidCount > 0 the tile has been invalidated, however the new tile content
+					// has not yet been fetched and because of `canceltiles` message it will never be
+					// so we need to remove the tile, or when the tile is back inside the visible area
+					// its content would be the old invalidated one;
+					// example: a tile is invalidated but a sudden scroll to the cell cursor position causes
+					// to move the tile out of the visible area before the new content is fetched
+					if (!tile.loaded || tile._invalidCount > 0) {
 						L.DomUtil.remove(tile.el);
 						delete this._tiles[key];
 						if (this._debug && this._debugDataCancelledTiles) {
@@ -684,8 +699,8 @@ L.GridLayer = L.Layer.extend({
 			if (tilePositionsX !== '' && tilePositionsY !== '') {
 				var message = 'tilecombine ' +
 					'part=' + this._selectedPart + ' ' +
-					'width=' + this._tileSize + ' ' +
-					'height=' + this._tileSize + ' ' +
+					'width=' + this._tileWidthPx + ' ' +
+					'height=' + this._tileHeightPx + ' ' +
 					'tileposx=' + tilePositionsX + ' ' +
 					'tileposy=' + tilePositionsY + ' ' +
 					'tilewidth=' + this._tileWidthTwips + ' ' +
@@ -747,8 +762,7 @@ L.GridLayer = L.Layer.extend({
 		if (!tile) { return; }
 
 		// FIXME: this _tileCache is used for prev/next slide; but it is
-		// dangerous in connection with typing / invalidation, so let's
-		// comment it out for now
+		// dangerous in connection with typing / invalidation
 		if (!(this._tiles[key]._invalidCount > 0)) {
 			this._tileCache[key] = tile.el.src;
 		}
@@ -916,8 +930,8 @@ L.GridLayer = L.Layer.extend({
 				twips = this._coordsToTwips(coords);
 				msg = 'tile ' +
 					'part=' + coords.part + ' ' +
-					'width=' + this._tileSize + ' ' +
-					'height=' + this._tileSize + ' ' +
+					'width=' + this._tileWidthPx + ' ' +
+					'height=' + this._tileHeightPx + ' ' +
 					'tileposx=' + twips.x + ' '	+
 					'tileposy=' + twips.y + ' ' +
 					'tilewidth=' + this._tileWidthTwips + ' ' +
@@ -946,8 +960,8 @@ L.GridLayer = L.Layer.extend({
 				twips = this._coordsToTwips(coords);
 				msg = 'tilecombine ' +
 					'part=' + coords.part + ' ' +
-					'width=' + this._tileSize + ' ' +
-					'height=' + this._tileSize + ' ' +
+					'width=' + this._tileWidthPx + ' ' +
+					'height=' + this._tileHeightPx + ' ' +
 					'tileposx=' + tilePositionsX + ' '	+
 					'tileposy=' + tilePositionsY + ' ' +
 					'tilewidth=' + this._tileWidthTwips + ' ' +
@@ -980,13 +994,13 @@ L.GridLayer = L.Layer.extend({
 			this._fadeFrame = L.Util.requestAnimFrame(this._updateOpacity, this);
 		} else {
 			tile.active = true;
-			this._pruneTiles();
 		}
 
 		L.DomUtil.addClass(tile.el, 'leaflet-tile-loaded');
 
 		if (this._noTilesToLoad()) {
 			this.fire('load');
+			this._pruneTiles();
 		}
 	},
 
