@@ -1,3 +1,4 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.TileLayer is used for standard xyz-numbered tile layers.
  */
@@ -45,10 +46,11 @@ L.TileLayer = L.GridLayer.extend({
 		maxNativeZoom: null, // Number
 		tms: false,
 		zoomReverse: false,
-		detectRetina: false,
+		detectRetina: true,
 		crossOrigin: false,
 		previewInvalidationTimeout: 1000,
-		defaultPermission: 'view'
+		marginX: 10,
+		marginY: 10
 	},
 
 	initialize: function (url, options) {
@@ -57,10 +59,13 @@ L.TileLayer = L.GridLayer.extend({
 
 		options = L.setOptions(this, options);
 
-		// detecting retina displays, adjusting tileSize and zoom levels
-		if (options.detectRetina && L.Browser.retina && options.maxZoom > 0) {
+		this._tileWidthPx = options.tileSize;
+		this._tileHeightPx = options.tileSize;
 
-			options.tileSize = Math.floor(options.tileSize / 2);
+		// detecting retina displays, adjusting tileWidthPx, tileHeightPx and zoom levels
+		if (options.docType !== 'spreadsheet' && options.detectRetina && L.Browser.retina && options.maxZoom > 0) {
+			this._tileWidthPx *= 2;
+			this._tileHeightPx *= 2;
 			options.zoomOffset++;
 
 			options.minZoom = Math.max(0, options.minZoom);
@@ -132,10 +137,8 @@ L.TileLayer = L.GridLayer.extend({
 		this._msgQueue = [];
 		this._toolbarCommandValues = {};
 		this._previewInvalidations = [];
-		this._partPageRectanglesTwips = [];
-		this._partPageRectanglesPixels = [];
-		this._clientZoom = 'tilepixelwidth=' + this.options.tileSize + ' ' +
-			'tilepixelheight=' + this.options.tileSize + ' ' +
+		this._clientZoom = 'tilepixelwidth=' + this._tileWidthPx + ' ' +
+			'tilepixelheight=' + this._tileHeightPx + ' ' +
 			'tiletwipwidth=' + this.options.tileWidthTwips + ' ' +
 			'tiletwipheight=' + this.options.tileHeightTwips;
 
@@ -168,6 +171,63 @@ L.TileLayer = L.GridLayer.extend({
 		this._levels = {};
 		this._tiles = {};
 		this._tileCache = {};
+		this._map._socket.sendMessage('commandvalues command=.uno:ViewAnnotations');
+		var that = this;
+		$.contextMenu({
+			selector: '.loleaflet-annotation-menu',
+			trigger: 'none',
+			className: 'loleaflet-font',
+			items: {
+				modify: {
+					name: _('Modify'),
+					callback: function (key, options) {
+						that.onAnnotationModify.call(that, options.$trigger.get(0).annotation);
+					}
+				},
+				reply: (this._docType !== 'text') ? undefined : {
+					name: _('Reply'),
+					callback: function (key, options) {
+						that.onAnnotationReply.call(that, options.$trigger.get(0).annotation);
+					}
+				},
+				remove: {
+					name: _('Remove'),
+					callback: function (key, options) {
+						that.onAnnotationRemove.call(that, options.$trigger.get(0).annotation._data.id);
+					}
+				}
+			},
+			events: {
+				show: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = true;
+				},
+				hide: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = false;
+				}
+			}
+		});
+		$.contextMenu({
+			selector: '.loleaflet-annotation-menu-redline',
+			trigger: 'none',
+			className: 'loleaflet-font',
+			items: {
+				modify: {
+					name: _('Comment'),
+					callback: function (key, options) {
+						that.onAnnotationModify.call(that, options.$trigger.get(0).annotation);
+					}
+				}
+			},
+			events: {
+				show: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = true;
+				},
+				hide: function (options) {
+					options.$trigger.get(0).annotation._contextMenu = false;
+				}
+			}
+		});
+		this._map._socket.sendMessage('commandvalues command=.uno:AcceptTrackedChanges');
 
 		map._fadeAnimated = false;
 		this._viewReset();
@@ -180,7 +240,6 @@ L.TileLayer = L.GridLayer.extend({
 		map.on('drop', this._onDrop, this);
 
 		map.on('zoomend', this._onUpdateCursor, this);
-		map.on('zoomend', this._onUpdatePartPageRectangles, this);
 		if (this._docType === 'spreadsheet') {
 			map.on('zoomend', this._onCellCursorShift, this);
 		}
@@ -305,6 +364,9 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('cellcursor:')) {
 			this._onCellCursorMsg(textMsg);
 		}
+		else if (textMsg.startsWith('celladdress:')) {
+			this._onCellAddressMsg(textMsg);
+		}
 		else if (textMsg.startsWith('cellformula:')) {
 			this._onCellFormulaMsg(textMsg);
 		}
@@ -336,9 +398,6 @@ L.TileLayer = L.GridLayer.extend({
 		else if (textMsg.startsWith('mousepointer:')) {
 			this._onMousePointerMsg(textMsg);
 		}
-		else if (textMsg.startsWith('partpagerectangles:')) {
-			this._onPartPageRectanglesMsg(textMsg);
-		}
 		else if (textMsg.startsWith('renderfont:')) {
 			this._onRenderFontMsg(textMsg, img);
 		}
@@ -427,13 +486,15 @@ L.TileLayer = L.GridLayer.extend({
 		}
 		else if (obj.commandName === '.uno:CellCursor') {
 			this._onCellCursorMsg(obj.commandValues);
-		} else if (this._map.unoToolbarCommands.indexOf(obj.commandName) !== -1) {
+		}
+		else if (this._map.unoToolbarCommands.indexOf(obj.commandName) !== -1) {
 			this._toolbarCommandValues[obj.commandName] = obj.commandValues;
 			this._map.fire('updatetoolbarcommandvalues', {
 				commandName: obj.commandName,
 				commandValues: obj.commandValues
 			});
-		} else {
+		}
+		else {
 			this._map.fire('commandvalues', {
 				commandName: obj.commandName,
 				commandValues: obj.commandValues
@@ -441,6 +502,11 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
+	_onCellAddressMsg: function (textMsg) {
+		var address = textMsg.substring(13);
+		this._map.fire('celladdress', {address: address});
+	},
+
 	_onCellFormulaMsg: function (textMsg) {
 		var formula = textMsg.substring(13);
 		this._selectionTextContent = formula;
@@ -463,8 +529,7 @@ L.TileLayer = L.GridLayer.extend({
 
 		this._map.hideBusy();
 		if (command.id === 'print') {
-			var isFirefox = typeof InstallTrigger !== 'undefined' || navigator.userAgent.search('Firefox') >= 0;
-			if (isFirefox || this._map.options.print === false) {
+			if (L.Browser.gecko || L.Browser.edge || this._map.options.print === false) {
 				// the print dialog doesn't work well on firefox
 				this._map.fire('print', {url: url});
 			}
@@ -475,7 +540,7 @@ L.TileLayer = L.GridLayer.extend({
 		else if (command.id === 'slideshow') {
 			this._map.fire('slidedownloadready', {url: url});
 		}
-		else {
+		else if (command.id === 'export') {
 			this._map._fileDownloader.src = url;
 		}
 	},
@@ -559,6 +624,13 @@ L.TileLayer = L.GridLayer.extend({
 		if (!this._prevCellCursor) {
 			this._prevCellCursor = L.LatLngBounds.createDefault();
 		}
+		if (!this._cellCursorXY) {
+			this._cellCursorXY = new L.Point(-1, -1);
+		}
+		if (!this._prevCellCursorXY) {
+			this._prevCellCursorXY = new L.Point(-1, -1);
+		}
+
 		if (textMsg.match('EMPTY')) {
 			this._cellCursorTwips = new L.Bounds(new L.Point(0, 0), new L.Point(0, 0));
 			this._cellCursor = L.LatLngBounds.createDefault();
@@ -572,6 +644,7 @@ L.TileLayer = L.GridLayer.extend({
 			this._cellCursor = new L.LatLngBounds(
 							this._twipsToLatLng(topLeftTwips, this._map.getZoom()),
 							this._twipsToLatLng(bottomRightTwips, this._map.getZoom()));
+			this._cellCursorXY = new L.Point(parseInt(strTwips[4]), parseInt(strTwips[5]));
 		}
 
 		var horizontalDirection = 0;
@@ -716,10 +789,10 @@ L.TileLayer = L.GridLayer.extend({
 
 		if (!this._isEmptyRectangle(this._cellViewCursors[viewId].bounds) && this._selectedPart === viewPart) {
 			if (!cellViewCursorMarker) {
-				var borderColor = L.LOUtil.rgbToHex(this._map.getViewColor(viewId));
-				cellViewCursorMarker = L.rectangle(this._cellViewCursors[viewId].bounds, {fill: false, color: borderColor, weight: 2});
+				var backgroundColor = L.LOUtil.rgbToHex(this._map.getViewColor(viewId));
+				cellViewCursorMarker = L.rectangle(this._cellViewCursors[viewId].bounds, {fill: false, color: backgroundColor, weight: 2});
 				this._cellViewCursors[viewId].marker = cellViewCursorMarker;
-				cellViewCursorMarker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, borderColor: borderColor});
+				cellViewCursorMarker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, backgroundColor: backgroundColor, color: 'white', closeButton: false});
 			}
 			else {
 				cellViewCursorMarker.setBounds(this._cellViewCursors[viewId].bounds);
@@ -763,8 +836,8 @@ L.TileLayer = L.GridLayer.extend({
 				}
 			}
 
-			var borderColor = L.LOUtil.rgbToHex(this._map.getViewColor(viewId));
-			this._cellViewCursors[viewId].marker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, borderColor: borderColor});
+			var backgroundColor = L.LOUtil.rgbToHex(this._map.getViewColor(viewId));
+			this._cellViewCursors[viewId].marker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, backgroundColor: backgroundColor, color: 'white', closeButton: false});
 		}
 	},
 
@@ -785,19 +858,19 @@ L.TileLayer = L.GridLayer.extend({
 		this._onUpdateViewCursor(viewId);
 	},
 
-	_addView: function(viewId, userid, username, color) {
-		if (color === 0 && this._map.getDocType() !== 'text') {
-			color = L.LOUtil.getViewIdColor(viewId);
+	_addView: function(viewInfo) {
+		if (viewInfo.color === 0 && this._map.getDocType() !== 'text') {
+			viewInfo.color = L.LOUtil.getViewIdColor(viewInfo.id);
 		}
 
-		this._map.addView(viewId, userid, username, color);
+		this._map.addView(viewInfo);
 
 		//TODO: We can initialize color and other properties here.
-		if (typeof this._viewCursors[viewId] !== 'undefined') {
-			this._viewCursors[viewId] = {};
+		if (typeof this._viewCursors[viewInfo.id] !== 'undefined') {
+			this._viewCursors[viewInfo.id] = {};
 		}
 
-		this._onUpdateViewCursor(viewId);
+		this._onUpdateViewCursor(viewInfo.id);
 	},
 
 	_removeView: function(viewId) {
@@ -831,7 +904,7 @@ L.TileLayer = L.GridLayer.extend({
 		var viewIds = [];
 		for (var viewInfoIdx in viewInfo) {
 			if (!(parseInt(viewInfo[viewInfoIdx].id) in this._map._viewInfo)) {
-				this._addView(viewInfo[viewInfoIdx].id, viewInfo[viewInfoIdx].userid, viewInfo[viewInfoIdx].username, viewInfo[viewInfoIdx].color);
+				this._addView(viewInfo[viewInfoIdx]);
 			}
 			viewIds.push(viewInfo[viewInfoIdx].id);
 		}
@@ -844,49 +917,24 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
-	_onPartPageRectanglesMsg: function (textMsg) {
-		textMsg = textMsg.substring(19);
-		var pages = textMsg.split(';');
-		this._partPageRectanglesTwips = [];
-		this._partPageRectanglesPixels = [];
-		for (var i = 0; i < pages.length; i++) {
-			var strTwips = pages[i].match(/\d+/g);
-			if (!strTwips) {
-				// probably not a text file
-				return;
-			}
-			var topLeftTwips = new L.Point(parseInt(strTwips[0]), parseInt(strTwips[1]));
-			var offset = new L.Point(parseInt(strTwips[2]), parseInt(strTwips[3]));
-			var bottomRightTwips = topLeftTwips.add(offset);
-			var pageBoundsTwips = new L.Bounds(topLeftTwips, bottomRightTwips);
-			this._partPageRectanglesTwips.push(pageBoundsTwips);
-			var pageBoundsPixels = new L.Bounds(
-					this._twipsToPixels(topLeftTwips),
-					this._twipsToPixels(bottomRightTwips));
-			this._partPageRectanglesPixels.push(pageBoundsPixels);
-		}
-		this._map.fire('partpagerectangles', {
-			pixelRectangles: this._partPageRectanglesPixels,
-			twipsRectangles: this._partPageRectanglesTwips
-		});
-		this._onCurrentPageUpdate();
-	},
-
 	_onRenderFontMsg: function (textMsg, img) {
 		var command = this._map._socket.parseServerCmd(textMsg);
 		this._map.fire('renderfont', {
 			font: command.font,
+			char: command.char,
 			img: img
 		});
 	},
 
 	_onSearchNotFoundMsg: function (textMsg) {
 		this._clearSearchResults();
+		this._searchRequested = false;
 		var originalPhrase = textMsg.substring(16);
 		this._map.fire('search', {originalPhrase: originalPhrase, count: 0});
 	},
 
 	_onSearchResultSelection: function (textMsg) {
+		this._searchRequested = false;
 		textMsg = textMsg.substring(23);
 		var obj = JSON.parse(textMsg);
 		var originalPhrase = obj.searchString;
@@ -907,12 +955,15 @@ L.TileLayer = L.GridLayer.extend({
 			this._clearSearchResults();
 			this._searchResults = results;
 			this._map.setPart(results[0].part); // go to first result.
+		} else if (count === 1) {
+			this._lastSearchResult = results[0];
 		}
 		this._searchTerm = originalPhrase;
 		this._map.fire('search', {originalPhrase: originalPhrase, count: count, highlightAll: highlightAll, results: results});
 	},
 
 	_clearSearchResults: function() {
+		this._lastSearchResult = null;
 		this._searchResults = null;
 		this._searchTerm = null;
 		this._searchResultsLayer.clearLayers();
@@ -972,8 +1023,7 @@ L.TileLayer = L.GridLayer.extend({
 		}
 
 		this._map.hideBusy();
-		// TODO when implemented in the LOK, add also obj.result
-		this._map.fire('commandresult', {commandName: commandName, success: success});
+		this._map.fire('commandresult', {commandName: commandName, success: success, result: obj.result});
 
 	},
 
@@ -1133,6 +1183,9 @@ L.TileLayer = L.GridLayer.extend({
 			});
 		}
 		else if (tile) {
+			if (command.hash != undefined) {
+				tile.oldhash = command.hash;
+			}
 			if (this._tiles[key]._invalidCount > 0) {
 				this._tiles[key]._invalidCount -= 1;
 			}
@@ -1144,9 +1197,6 @@ L.TileLayer = L.GridLayer.extend({
 			}
 			tile.el.src = img;
 		}
-		else if (command.preFetch === 'true') {
-			this._tileCache[key] = img;
-		}
 		L.Log.log(textMsg, L.INCOMING, key);
 	},
 
@@ -1217,8 +1267,7 @@ L.TileLayer = L.GridLayer.extend({
 	},
 
 	_postKeyboardEvent: function(type, charcode, keycode) {
-		if (this._docType === 'spreadsheet' && this._prevCellCursor && type === 'input')
-		{
+		if (this._docType === 'spreadsheet' && this._prevCellCursor && type === 'input') {
 			if (keycode === 1030) { // PgUp
 				if (this._cellCursorOnPgUp) {
 					return;
@@ -1271,8 +1320,6 @@ L.TileLayer = L.GridLayer.extend({
 
 	// Update cursor layer (blinking cursor).
 	_onUpdateCursor: function (e) {
-		var pixBounds = L.bounds(this._map.latLngToLayerPoint(this._visibleCursor.getSouthWest()),
-						 this._map.latLngToLayerPoint(this._visibleCursor.getNorthEast()));
 		var cursorPos = this._visibleCursor.getNorthWest();
 
 		if (!e && !this._map.getBounds().contains(this._visibleCursor) && this._isCursorVisible) {
@@ -1287,11 +1334,31 @@ L.TileLayer = L.GridLayer.extend({
 			}
 		}
 
-		if (this._map._permission === 'edit' && this._isCursorVisible && this._isCursorOverlayVisible
-				&& !this._isEmptyRectangle(this._visibleCursor)) {
+		this._updateCursorAndOverlay();
+
+		this.eachView(this._viewCursors, function (item) {
+			var viewCursorMarker = item.marker;
+			if (viewCursorMarker) {
+				viewCursorMarker.setOpacity(this._map.hasLayer(this._cursorMarker) && this._cursorMarker.getLatLng().equals(viewCursorMarker.getLatLng()) ? 0 : 1);
+			}
+		}, this, true);
+	},
+
+	// enable or disable blinking cursor and  the cursor overlay depending on
+	// the state of the document (if the falgs are set)
+	_updateCursorAndOverlay: function (update) {
+		if (this._map._permission === 'edit'
+		&& this._isCursorVisible
+		&& this._isCursorOverlayVisible
+		&& !this._isEmptyRectangle(this._visibleCursor)) {
+
+			var pixBounds = L.bounds(this._map.latLngToLayerPoint(this._visibleCursor.getSouthWest()),
+									 this._map.latLngToLayerPoint(this._visibleCursor.getNorthEast()));
+
+			var cursorPos = this._visibleCursor.getNorthWest();
+
 			if (!this._cursorMarker) {
-				this._cursorMarker = L.cursor(cursorPos, pixBounds.getSize().multiplyBy(this._map.getZoomScale(this._map.getZoom())),
-					{blink: true});
+				this._cursorMarker = L.cursor(cursorPos, pixBounds.getSize().multiplyBy(this._map.getZoomScale(this._map.getZoom())), {blink: true});
 			}
 			else {
 				this._cursorMarker.setLatLng(cursorPos, pixBounds.getSize().multiplyBy(this._map.getZoomScale(this._map.getZoom())));
@@ -1302,13 +1369,6 @@ L.TileLayer = L.GridLayer.extend({
 			this._map.removeLayer(this._cursorMarker);
 			this._isCursorOverlayVisible = false;
 		}
-
-		this.eachView(this._viewCursors, function (item) {
-			var viewCursorMarker = item.marker;
-			if (viewCursorMarker) {
-				viewCursorMarker.setOpacity(this._map.hasLayer(this._cursorMarker) && this._cursorMarker.getLatLng().equals(viewCursorMarker.getLatLng()) ? 0 : 1);
-			}
-		}, this, true);
 	},
 
 	// Update colored non-blinking view cursor
@@ -1325,7 +1385,7 @@ L.TileLayer = L.GridLayer.extend({
 		var viewCursorVisible = this._viewCursors[viewId].visible;
 		var viewPart = this._viewCursors[viewId].part;
 
-		if (viewCursorVisible && !this._isEmptyRectangle(this._viewCursors[viewId].bounds) &&
+		if (!this._map.isViewReadOnly(viewId) && viewCursorVisible && !this._isEmptyRectangle(this._viewCursors[viewId].bounds) &&
 		   (this._docType === 'text' || this._selectedPart === viewPart)) {
 			if (!viewCursorMarker) {
 				var viewCursorOptions = {
@@ -1415,7 +1475,7 @@ L.TileLayer = L.GridLayer.extend({
 					color: color
 				});
 				// Disable autoPan, so the graphic view selection doesn't make the view jump to the popup.
-				viewMarker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, borderColor: color});
+				viewMarker.bindPopup(this._map.getViewName(viewId), {autoClose: false, autoPan: false, backgroundColor: color, color: 'white', closeButton: false});
 				this._graphicViewMarkers[viewId].marker = viewMarker;
 			}
 			else {
@@ -1536,7 +1596,7 @@ L.TileLayer = L.GridLayer.extend({
 	_onUpdateCellCursor: function (horizontalDirection, verticalDirection, onPgUpDn) {
 		if (this._cellCursor && !this._isEmptyRectangle(this._cellCursor)) {
 			var mapBounds = this._map.getBounds();
-			if (!mapBounds.contains(this._cellCursor)) {
+			if (!mapBounds.contains(this._cellCursor) && !this._cellCursorXY.equals(this._prevCellCursorXY)) {
 				var scrollX = 0, scrollY = 0;
 				if (onPgUpDn) {
 					var mapHalfHeight = (mapBounds.getNorth() - mapBounds.getSouth()) / 2;
@@ -1549,17 +1609,18 @@ L.TileLayer = L.GridLayer.extend({
 						scrollY = (this._cellCursor.getNorth() - mapBounds.getNorth()) + mapHalfHeight;
 					}
 				}
-				else {
+				else if (horizontalDirection !== 0 || verticalDirection != 0) {
 					var spacingX = Math.abs(this._cellCursor.getEast() - this._cellCursor.getWest()) / 4.0;
 					var spacingY = Math.abs((this._cellCursor.getSouth() - this._cellCursor.getNorth())) / 4.0;
-					if (horizontalDirection === -1 && this._cellCursor.getWest() < mapBounds.getWest()) {
+
+					if (this._cellCursor.getWest() < mapBounds.getWest()) {
 						scrollX = this._cellCursor.getWest() - mapBounds.getWest() - spacingX;
-					} else if (horizontalDirection === 1 && this._cellCursor.getEast() > mapBounds.getEast()) {
+					} else if (this._cellCursor.getEast() > mapBounds.getEast()) {
 						scrollX = this._cellCursor.getEast() - mapBounds.getEast() + spacingX;
 					}
-					if (verticalDirection === 1 && this._cellCursor.getNorth() > mapBounds.getNorth()) {
+					if (this._cellCursor.getNorth() > mapBounds.getNorth()) {
 						scrollY = this._cellCursor.getNorth() - mapBounds.getNorth() + spacingY;
-					} else if (verticalDirection === -1 && this._cellCursor.getSouth() < mapBounds.getSouth()) {
+					} else if (this._cellCursor.getSouth() < mapBounds.getSouth()) {
 						scrollY = this._cellCursor.getSouth() - mapBounds.getSouth() - spacingY;
 					}
 				}
@@ -1573,6 +1634,7 @@ L.TileLayer = L.GridLayer.extend({
 					center.y = Math.round(center.y < 0 ? 0 : center.y);
 					this._map.fire('scrollto', {x: center.x, y: center.y});
 				}
+				this._prevCellCursorXY = this._cellCursorXY;
 			}
 
 			if (onPgUpDn) {
@@ -1674,8 +1736,9 @@ L.TileLayer = L.GridLayer.extend({
 
 			// remember the copied text, for rich copy/paste inside a document
 			this._selectionTextHash = this._selectionTextContent;
-			this._map._socket.sendMessage('uno .uno:Copy');
 		}
+
+		this._map._socket.sendMessage('uno .uno:Copy');
 	},
 
 	_onCut: function (e) {
@@ -1686,23 +1749,24 @@ L.TileLayer = L.GridLayer.extend({
 
 			// remember the copied text, for rich copy/paste inside a document
 			this._selectionTextHash = this._selectionTextContent;
-			this._map._socket.sendMessage('uno .uno:Cut');
 		}
+
+		this._map._socket.sendMessage('uno .uno:Cut');
 	},
 
 	_onPaste: function (e) {
 		e = e.originalEvent;
 		e.preventDefault();
 		var pasteString = L.Compatibility.clipboardGet(e);
-		if (pasteString) {
-			if (pasteString === this._selectionTextHash) {
-				// content of the clipboard did not change, we can do rich
-				// paste
-				this._map._socket.sendMessage('uno .uno:Paste');
-			}
-			else {
-				this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + pasteString);
-			}
+		if (pasteString === 'false' || !pasteString || pasteString === this._selectionTextHash) {
+			// If there is nothing to paste in clipboard, no harm in
+			// issuing a .uno:Paste in case there is something internally copied in the document
+			// or if the content of the clipboard did not change, we surely must do a rich paste
+			// instead of a normal paste
+			this._map._socket.sendMessage('uno .uno:Paste');
+		}
+		else {
+			this._map._socket.sendMessage('paste mimetype=text/plain;charset=utf-8\n' + pasteString);
 		}
 	},
 
@@ -1779,7 +1843,7 @@ L.TileLayer = L.GridLayer.extend({
 	_fitWidthZoom: function (e, maxZoom) {
 		var size = e ? e.newSize : this._map.getSize();
 		var widthTwips = size.x * this._map.options.tileWidthTwips / this._tileSize;
-		maxZoom = maxZoom ? maxZoom : this._map.options.zoom;
+		maxZoom = maxZoom ? maxZoom : this._map.getZoom();
 
 		// 'fit width zoom' has no use in spreadsheets, ignore it there
 		if (this._docType !== 'spreadsheet') {
@@ -1787,7 +1851,7 @@ L.TileLayer = L.GridLayer.extend({
 			if (this._docWidthTwips > 0)
 			{
 				var ratio = widthTwips / this._docWidthTwips;
-				var zoom = this._map.options.zoom + Math.floor(Math.log(ratio) / Math.log(crsScale));
+				var zoom = this._map.getZoom() + Math.floor(Math.log(ratio) / Math.log(crsScale));
 
 				zoom = Math.max(1, zoom);
 				zoom = Math.min(maxZoom, zoom);
@@ -1816,29 +1880,13 @@ L.TileLayer = L.GridLayer.extend({
 		}
 	},
 
-	_onUpdatePartPageRectangles: function () {
-		if (this._partPageRectanglesPixels.length > 0) {
-			this._partPageRectanglesPixels = [];
-			for (var i = 0; i < this._partPageRectanglesTwips.length; i++) {
-				var pageBounds = new L.Bounds(
-						this._twipsToPixels(this._partPageRectanglesTwips[i].min),
-						this._twipsToPixels(this._partPageRectanglesTwips[i].max));
-				this._partPageRectanglesPixels.push(pageBounds);
-			}
-			this._map.fire('partpagerectangles', {
-				pixelRectangles: this._partPageRectanglesPixels,
-				twipsRectangles: this._partPageRectanglesTwips
-			});
-		}
-	},
-
 	// Cells can change position during changes of zoom level in calc
 	// hence we need to request an updated cell cursor position for this level.
 	_onCellCursorShift: function (force) {
 		if (this._cellCursorMarker || force) {
 			this._map._socket.sendMessage('commandvalues command=.uno:CellCursor'
-			                     + '?outputHeight=' + this._tileSize
-			                     + '&outputWidth=' + this._tileSize
+			                     + '?outputHeight=' + this._tileWidthPx
+			                     + '&outputWidth=' + this._tileHeightPx
 			                     + '&tileHeight=' + this._tileWidthTwips
 			                     + '&tileWidth=' + this._tileHeightTwips);
 		}
@@ -1910,8 +1958,8 @@ L.TileLayer = L.GridLayer.extend({
 	},
 
 	_updateClientZoom: function () {
-		this._clientZoom = 'tilepixelwidth=' + this._tileSize + ' ' +
-			'tilepixelheight=' + this._tileSize + ' ' +
+		this._clientZoom = 'tilepixelwidth=' + this._tileWidthPx + ' ' +
+			'tilepixelheight=' + this._tileHeightPx + ' ' +
 			'tiletwipwidth=' + this._tileWidthTwips + ' ' +
 			'tiletwipheight=' + this._tileHeightTwips;
 	},
@@ -1955,18 +2003,22 @@ L.TileLayer = L.GridLayer.extend({
 			}
 			this._debugInfo = new L.LayerGroup();
 			this._debugInfo2 = new L.LayerGroup();
+			this._debugAlwaysActive = new L.LayerGroup();
 			this._debugTyper = new L.LayerGroup();
 			map.addLayer(this._debugInfo);
 			map.addLayer(this._debugInfo2);
 			var overlayMaps = {
 				'Tile overlays': this._debugInfo,
 				'Screen overlays': this._debugInfo2,
+				'Always active': this._debugAlwaysActive,
 				'Typing': this._debugTyper
 			};
 			L.control.layers({}, overlayMaps, {collapsed: false}).addTo(map);
 
 			this._map.on('layeradd', function(e) {
-				if (e.layer === this._debugTyper) {
+				if (e.layer === this._debugAlwaysActive) {
+					map._debugAlwaysActive = true;
+				} else if (e.layer === this._debugTyper) {
 					this._debugTypeTimeout();
 				} else if (e.layer === this._debugInfo2) {
 					for (var i = 0; i < this._debugDataNames.length; i++) {
@@ -1975,7 +2027,9 @@ L.TileLayer = L.GridLayer.extend({
 				}
 			}, this);
 			map.on('layerremove', function(e) {
-				if (e.layer === this._debugTyper) {
+				if (e.layer === this._debugAlwaysActive) {
+					map._debugAlwaysActive = false;
+				} else if (e.layer === this._debugTyper) {
 					clearTimeout(this._debugTypeTimeoutId);
 				} else if (e.layer === this._debugInfo2) {
 					for (var i in this._debugData) {
@@ -2032,6 +2086,7 @@ L.TileLayer = L.GridLayer.extend({
 		if (tile._debugTile) {
 			tile._debugTile.setStyle({fillOpacity: 0.5, fillColor: 'blue'});
 			tile._debugTime.date = +new Date();
+			tile._debugTile.date = +new Date();
 			tile._debugInvalidateCount++;
 			this._debugInvalidateCount++;
 		}
@@ -2066,6 +2121,14 @@ L.TileLayer = L.GridLayer.extend({
 					rect.setStyle({fillOpacity: opac - 0.04});
 				}
 			}
+			for (var key in this._debugTiles) {
+				var rect = this._debugTiles[key];
+				var col = rect.options.fillColor;
+				var opac = rect.options.fillOpacity;
+				if (col === 'blue' && opac >= 0.04 && rect.date + 1000 < +new Date()) {
+					rect.setStyle({fillOpacity: opac - 0.04});
+				}
+			}
 			this._debugTimeoutId = setTimeout(function () { map._docLayer._debugTimeout(); }, 50);
 		}
 	},
