@@ -1,8 +1,9 @@
+/* -*- js-indent-level: 8 -*- */
 /*
  * L.Map is the central class of the API - it is used to create a map.
  */
 
-/* global vex $ w2ui w2utils */
+/* global vex $ */
 L.Map = L.Evented.extend({
 
 	options: {
@@ -20,6 +21,8 @@ L.Map = L.Evented.extend({
 		urlPrefix: 'lool'
 	},
 
+	lastActiveTime: Date.now(),
+
 	initialize: function (id, options) { // (HTMLElement or String, Object)
 		options = L.setOptions(this, options);
 
@@ -67,6 +70,8 @@ L.Map = L.Evented.extend({
 		this._active = true;
 		this._fatal = false;
 		this._enabled = true;
+		this._debugAlwaysActive = false; // disables the dimming / document inactivity when true
+		this._serverRecycling = false;
 
 		vex.dialogID = -1;
 
@@ -93,11 +98,20 @@ L.Map = L.Evented.extend({
 		};
 		this.initComplete = false;
 
-		this.on('updatepermission', function() {
+		this.on('updatepermission', function(e) {
 			if (!this.initComplete) {
 				this._fireInitComplete('updatepermission');
 			}
-		});
+
+			if (e.perm === 'readonly') {
+				L.DomUtil.addClass(this._container.parentElement, 'readonly');
+				L.DomUtil.addClass(L.DomUtil.get('logo'), 'readonly');
+				L.DomUtil.addClass(L.DomUtil.get('toolbar-wrapper'), 'readonly');
+				L.DomUtil.addClass(L.DomUtil.get('main-menu'), 'readonly');
+				L.DomUtil.addClass(L.DomUtil.get('presentation-controls-wrapper'), 'readonly');
+				L.DomUtil.addClass(L.DomUtil.get('spreadsheet-row-column-frame'), 'readonly');
+			}
+		}, this);
 		this.on('doclayerinit', function() {
 			if (!this.initComplete) {
 				this._fireInitComplete('doclayerinit');
@@ -123,16 +137,37 @@ L.Map = L.Evented.extend({
 
 		// View color map
 		this._viewColors = {};
+
+		// This becomes true if document was ever modified by the user
+		this._everModified = false;
+
+		this.on('commandstatechanged', function(e) {
+			if (e.commandName === '.uno:ModifiedStatus')
+				this._everModified = this._everModified || (e.state === 'true');
+		}, this);
+	},
+
+	loadDocument: function() {
+		this._socket.connect();
 	},
 
 	// public methods that modify map state
 
-	addView: function(viewid, userid, username, color) {
-		this._viewInfo[viewid] = {'userid': userid, 'username': username, 'color': color};
-		this.fire('postMessage', {msgId: 'View_Added', args: {ViewId: viewid, UserId: userid, UserName: username, Color: color}});
+	getViewId: function (username) {
+		for (var idx in this._viewInfo) {
+			if (this._viewInfo[idx].username === username) {
+				return this._viewInfo[idx].id;
+			}
+		}
+		return -1;
+	},
+
+	addView: function(viewInfo) {
+		this._viewInfo[viewInfo.id] = viewInfo;
+		this.fire('postMessage', {msgId: 'View_Added', args: {ViewId: viewInfo.id, UserId: viewInfo.userid, UserName: viewInfo.username, Color: L.LOUtil.rgbToHex(viewInfo.color), ReadOnly: viewInfo.readonly}});
 
 		// Fire last, otherwise not all events are handled correctly.
-		this.fire('addview', {viewId: viewid, username: username});
+		this.fire('addview', {viewId: viewInfo.id, username: viewInfo.username, readonly: this.isViewReadOnly(viewInfo.id)});
 	},
 
 	removeView: function(viewid) {
@@ -156,7 +191,7 @@ L.Map = L.Evented.extend({
 		// If document is already loaded, ask the toolbar widget to show busy
 		// status on the bottom statusbar
 		if (this._docLayer) {
-			w2utils.lock(w2ui['toolbar-down'].box, label, true);
+			this.fire('showbusy', {label: label});
 			return;
 		}
 
@@ -170,9 +205,7 @@ L.Map = L.Evented.extend({
 	},
 
 	hideBusy: function () {
-		if (w2ui['toolbar-down'].box.firstChild.className === 'w2ui-lock') {
-			w2utils.unlock(w2ui['toolbar-down'].box);
-		}
+		this.fire('hidebusy');
 
 		if (this.hasLayer(this._progressBar)) {
 			this.removeLayer(this._progressBar);
@@ -250,10 +283,11 @@ L.Map = L.Evented.extend({
 		return this.fire('moveend');
 	},
 
-	setMaxBounds: function (bounds) {
+	setMaxBounds: function (bounds, options) {
 		bounds = L.latLngBounds(bounds);
 
 		this.options.maxBounds = bounds;
+		options = options || {};
 
 		if (!bounds) {
 			return this.off('moveend', this._panInsideMaxBounds);
@@ -263,6 +297,10 @@ L.Map = L.Evented.extend({
 			this._panInsideMaxBounds();
 		}
 
+		if (options.panInside === false) {
+			return this.off('moveend', this._panInsideMaxBounds);
+		}
+
 		return this.on('moveend', this._panInsideMaxBounds);
 	},
 
@@ -285,7 +323,6 @@ L.Map = L.Evented.extend({
 
 		var oldSize = this.getSize();
 		this._sizeChanged = true;
-		this._initialCenter = null;
 
 		var newSize = this.getSize(),
 		    oldCenter = oldSize.divideBy(2).round(),
@@ -393,12 +430,12 @@ L.Map = L.Evented.extend({
 		return this._viewInfo[viewid].color;
 	},
 
+	isViewReadOnly: function(viewid) {
+		return this._viewInfo[viewid].readonly !== '0';
+	},
+
 	getCenter: function () { // (Boolean) -> LatLng
 		this._checkIfLoaded();
-
-		if (this._initialCenter && !this._moved()) {
-			return this._initialCenter;
-		}
 		return this.layerPointToLatLng(this._getCenterLayerPoint());
 	},
 
@@ -509,7 +546,8 @@ L.Map = L.Evented.extend({
 
 	project: function (latlng, zoom) { // (LatLng[, Number]) -> Point
 		zoom = zoom === undefined ? this._zoom : zoom;
-		return this.options.crs.latLngToPoint(L.latLng(latlng), zoom);
+		var projectedPoint = this.options.crs.latLngToPoint(L.latLng(latlng), zoom);
+		return new L.Point(L.round(projectedPoint.x, 1e-6), L.round(projectedPoint.y, 1e-6));
 	},
 
 	unproject: function (point, zoom) { // (Point[, Number]) -> LatLng
@@ -565,7 +603,9 @@ L.Map = L.Evented.extend({
 	},
 
 	focus: function () {
+		console.debug('focus:');
 		if (this._docLayer && document.activeElement !== this._docLayer._textArea) {
+			console.debug('focus: focussing');
 			this._docLayer._textArea.focus();
 		}
 	},
@@ -595,7 +635,12 @@ L.Map = L.Evented.extend({
 		}
 
 		var textAreaContainer = L.DomUtil.create('div', 'clipboard-container', container.parentElement);
-		this._textArea = L.DomUtil.create('textarea', 'clipboard', textAreaContainer);
+		this._textArea = L.DomUtil.create('input', 'clipboard', textAreaContainer);
+		this._textArea.setAttribute('type', 'text');
+		this._textArea.setAttribute('autocorrect', 'off');
+		this._textArea.setAttribute('autocapitalize', 'off');
+		this._textArea.setAttribute('autocomplete', 'off');
+		this._textArea.setAttribute('spellcheck', 'false');
 		this._resizeDetector = L.DomUtil.create('iframe', 'resize-detector', container);
 		this._fileDownloader = L.DomUtil.create('iframe', '', container);
 		L.DomUtil.setStyle(this._fileDownloader, 'display', 'none');
@@ -662,7 +707,6 @@ L.Map = L.Evented.extend({
 		}
 
 		this._zoom = zoom;
-		this._initialCenter = center;
 
 		if (!preserveMapOffset) {
 			L.DomUtil.setPosition(this._mapPane, new L.Point(0, 0));
@@ -738,38 +782,127 @@ L.Map = L.Evented.extend({
 	},
 
 	_activate: function () {
+		if (this._serverRecycling) {
+			return;
+		}
+
+		console.debug('_activate:');
 		clearTimeout(vex.timer);
 
 		if (!this._active) {
 			// Only activate when we are connected.
 			if (this._socket.connected()) {
+				console.debug('sending useractive');
 				this._socket.sendMessage('useractive');
 				this._active = true;
-				this._docLayer._onMessage('invalidatetiles: EMPTY', null);
+				if (this._docLayer) {
+					this._docLayer._onMessage('invalidatetiles: EMPTY', null);
+				}
+				if (vex.dialogID > 0) {
+					var id = vex.dialogID;
+					vex.dialogID = -1;
+					this._startInactiveTimer();
+					this.focus();
+					return vex.close(id);
+				}
 			} else {
-				this._active = true;
 				this._socket.initialize(this);
 			}
-
-			if (vex.dialogID > 0) {
-				var id = vex.dialogID;
-				vex.dialogID = -1;
-				return vex.close(id);
-			}
 		}
 
+		this._startInactiveTimer();
+		this.focus();
 		return false;
 	},
 
-	_deactivate: function () {
+	_dim: function() {
+		if (this.options.alwaysActive || this._debugAlwaysActive === true) {
+			return;
+		}
+
+		console.debug('_dim:');
+		if (!this._socket.connected()) {
+			return;
+		}
+
+		this._active = false;
 		clearTimeout(vex.timer);
 
-		if (!window.devtools.open && (!this._active || vex.dialogID > 0)) {
+		var options = $.extend({}, vex.defaultOptions, {
+			contentCSS: {'background':'rgba(0, 0, 0, 0)',
+				     'font-size': 'xx-large',
+				     'color': '#fff',
+				     'text-align': 'center'},
+			content: _('Inactive document - please click to resume editing')
+		});
+		options.id = vex.globalID;
+		vex.dialogID = options.id;
+		vex.globalID += 1;
+		options.$vex = $('<div>').addClass(vex.baseClassNames.vex).addClass(options.className).css(options.css).data({
+			vex: options
+		});
+		options.$vexOverlay = $('<div>').addClass(vex.baseClassNames.overlay).addClass(options.overlayClassName).css(options.overlayCSS).data({
+			vex: options
+		});
+
+		var map = this;
+		options.$vex.bind('click.vex', function(e) {
+			console.debug('_dim: click.vex function');
+			return map._activate();
+		});
+		options.$vex.append(options.$vexOverlay);
+
+		options.$vexContent = $('<div>').addClass(vex.baseClassNames.content).addClass(options.contentClassName).css(options.contentCSS).text(options.content).data({
+			vex: options
+		});
+		options.$vex.append(options.$vexContent);
+
+		$(options.appendLocation).append(options.$vex);
+		vex.setupBodyClassName(options.$vex);
+
+		this._doclayer && this._docLayer._onMessage('textselection:', null);
+		console.debug('_dim: sending userinactive');
+		map.fire('postMessage', {msgId: 'User_Idle'});
+		this._socket.sendMessage('userinactive');
+	},
+
+	_dimIfInactive: function () {
+		console.debug('_dimIfInactive: diff=' + (Date.now() - this.lastActiveTime));
+		if ((Date.now() - this.lastActiveTime) >= this.options.idleTimeoutSecs * 1000) {
+			this._dim();
+		} else {
+			this._startInactiveTimer();
+		}
+	},
+
+	_startInactiveTimer: function () {
+		if (this._serverRecycling) {
+			return;
+		}
+
+		console.debug('_startInactiveTimer:');
+		clearTimeout(vex.timer);
+		var map = this;
+		vex.timer = setTimeout(function() {
+			map._dimIfInactive();
+		}, 1 * 60 * 1000); // Check once a minute
+	},
+
+	_deactivate: function () {
+		if (this._serverRecycling) {
+			return;
+		}
+
+		console.debug('_deactivate:');
+		clearTimeout(vex.timer);
+
+		if (!this._active || vex.dialogID > 0) {
 			// A dialog is already dimming the screen and probably
 			// shows an error message. Leave it alone.
 			this._active = false;
 			this._docLayer && this._docLayer._onMessage('textselection:', null);
 			if (this._socket.connected()) {
+				console.debug('_deactivate: sending userinactive');
 				this._socket.sendMessage('userinactive');
 			}
 
@@ -778,79 +911,46 @@ L.Map = L.Evented.extend({
 
 		var map = this;
 		vex.timer = setTimeout(function() {
-			if (window.devtools.open || !map._socket.connected()) {
-				return;
-			}
-
-			map._active = false;
-			clearTimeout(vex.timer);
-
-			var options = $.extend({}, vex.defaultOptions, {
-				contentCSS: {'background':'rgba(0, 0, 0, 0)',
-				             'font-size': 'xx-large',
-				             'color': '#fff',
-				             'text-align': 'center'},
-				content: _('Inactive document - please click to resume editing')
-			});
-			options.id = vex.globalID;
-			vex.dialogID = options.id;
-			vex.globalID += 1;
-			options.$vex = $('<div>').addClass(vex.baseClassNames.vex).addClass(options.className).css(options.css).data({
-				vex: options
-			});
-			options.$vexOverlay = $('<div>').addClass(vex.baseClassNames.overlay).addClass(options.overlayClassName).css(options.overlayCSS).data({
-				vex: options
-			});
-
-			options.$vexOverlay.bind('click.vex', function(e) {
-				if (e.target !== this) {
-					return 0;
-				}
-				return map._activate();
-			});
-			options.$vex.append(options.$vexOverlay);
-
-			options.$vexContent = $('<div>').addClass(vex.baseClassNames.content).addClass(options.contentClassName).css(options.contentCSS).text(options.content).data({
-				vex: options
-			});
-			options.$vex.append(options.$vexContent);
-
-			$(options.appendLocation).append(options.$vex);
-			vex.setupBodyClassName(options.$vex);
-
-			map._doclayer && map._docLayer._onMessage('textselection:', null);
-			map._socket.sendMessage('userinactive');
-
-		}, 30 * 1000); // Dim in 30 seconds.
+			map._dim();
+		}, map.options.outOfFocusTimeoutSecs * 1000);
 	},
 
 	_onLostFocus: function () {
 		if (!this._loaded) { return; }
 
+		console.debug('_onLostFocus: ');
 		var doclayer = this._docLayer;
-		if (doclayer && doclayer._isCursorVisible && doclayer._isCursorOverlayVisible) {
-			doclayer._visibleCursorOnLostFocus = doclayer._visibleCursor;
-			doclayer._isCursorOverlayVisibleOnLostFocus = doclayer._isCursorVisibleOnLostFocus = true;
+		if (!doclayer) { return; }
+
+		// save state of cursor (blinking marker) and the cursor overlay
+		doclayer._isCursorVisibleOnLostFocus = doclayer._isCursorVisible;
+		doclayer._isCursorOverlayVisibleOnLostFocus = doclayer._isCursorOverlayVisible;
+
+		// if the blinking cursor is visible, disable the overlay when we go out of focus
+		if (doclayer._isCursorVisible && doclayer._isCursorOverlayVisible) {
 			doclayer._isCursorOverlayVisible = false;
-			doclayer._onUpdateCursor();
+			doclayer._updateCursorAndOverlay();
 		}
 
 		this._deactivate();
 	},
 
 	_onGotFocus: function () {
+		console.debug('_onGotFocus:');
 		if (!this._loaded) { return; }
 
 		var doclayer = this._docLayer;
-		if (doclayer && doclayer._isCursorVisibleOnLostFocus && doclayer._isCursorOverlayVisibleOnLostFocus) {
+		if (doclayer &&
+		    typeof doclayer._isCursorOverlayVisibleOnLostFocus !== 'undefined' &&
+		    typeof doclayer._isCursorVisibleOnLostFocus !== 'undefined') {
 			// we restore the old cursor position by a small delay, so that if the user clicks
 			// inside the document we skip to restore it, so that the user does not see the cursor
 			// jumping from the old position to the new one
 			setTimeout(function () {
-				if (doclayer._isCursorOverlayVisible) { return; } // user has clicked inside the document
-				doclayer._isCursorOverlayVisible = doclayer._isCursorVisible = true;
-				doclayer._visibleCursor = doclayer._visibleCursorOnLostFocus;
-				doclayer._onUpdateCursor();
+				// restore the state that was before focus was lost
+				doclayer._isCursorOverlayVisible = doclayer._isCursorOverlayVisibleOnLostFocus;
+				doclayer._isCursorVisible = doclayer._isCursorVisibleOnLostFocus;
+				doclayer._updateCursorAndOverlay();
 			}, 300);
 		}
 
@@ -859,7 +959,7 @@ L.Map = L.Evented.extend({
 
 	_onUpdateProgress: function (e) {
 		if (e.statusType === 'start') {
-			if (this._socket.readyState === 1) {
+			if (this._socket.socket.readyState === 1) {
 				// auto-save
 				this.showBusy(_('Saving...'), true);
 			}
@@ -887,7 +987,9 @@ L.Map = L.Evented.extend({
 	},
 
 	_handleDOMEvent: function (e) {
-		if (!this._loaded || !this._enabled || L.DomEvent._skipped(e)) { return; }
+		if (!this._docLayer || !this._loaded || !this._enabled || L.DomEvent._skipped(e)) { return; }
+
+		this.lastActiveTime = Date.now();
 
 		// find the layer the event is propagating from
 		var target = this._targets[L.stamp(e.target || e.srcElement)],
@@ -898,8 +1000,15 @@ L.Map = L.Evented.extend({
 		// .focus() method on hidden input within actual 'click' event here
 		// Calling from some other place with no real 'click' event doesn't work
 		if (type === 'click') {
-			this._textArea.blur();
-			this._textArea.focus();
+			if (this._permission === 'edit') {
+				this._textArea.blur();
+				this._textArea.focus();
+			}
+
+			// unselect if anything is selected already
+			if (this._docLayer && this._docLayer._annotations && this._docLayer._annotations.unselect) {
+				this._docLayer._annotations.unselect();
+			}
 		}
 
 		// we need to keep track if we have entered/left the map
@@ -1059,6 +1168,11 @@ L.Map = L.Evented.extend({
 		return left + right > 0 ?
 			Math.round(left - right) / 2 :
 			Math.max(0, Math.ceil(left)) - Math.max(0, Math.floor(right));
+		// TODO: do we really need ceil and floor ?
+		// for spreadsheets it can cause one pixel alignment offset btw grid and row/column header
+		// and a one pixel horizontal auto-scrolling issue;
+		// both issues have been fixed by rounding the projection: see Map.project above;
+		// anyway in case of similar problems, this code needs to be checked
 	},
 
 	_limitZoom: function (zoom) {
